import time
import os

import requests
import xml.etree.ElementTree


class DatabricksApiException(Exception):
    """Exception class for Databricks API errors
    
    Args:
        status_code (int): HTTP status code
        error_code (int): Error specific code
        error_message (str): Error message

    Attributes:
        status_code (int): HTTP status code
        error_code (int): Error specific code
        error_message (str): Error message
    """
    def __init__(self, status_code, error_code, error_message):
        self.status_code = status_code
        self.error_code = error_code
        self.error_message = error_message


class Rest(object):
    """Rest class to execute REST get and post calls.
    JSON Results will automatically be converted to dicts
    """
    def _rest_error(self, status_code, error_code, message):
        """Create a stabdard REST error message
        
        Args:
            status_code (str): HTTP status code
            error_code (str): Error code
            message (str): Error message
        
        Returns:
            dict: Dictionary with status_code, error_code and message
        """
        return {"status_code": status_code, "error_code": error_code, "message": message}

    def _remove_tags(self, text):
        """Very simple routine to get rid of HTML tags of error responses
        
        Args:
            text (str): HTML error message
        
        Returns:
            str: The error message without HTML tags
        """
        try:
            result = ''.join(xml.etree.ElementTree.fromstring(text).itertext()).replace("\n\n", "\n")
        except:
            result = text
        return result

    def _json(self, response):
        """Convert a JSON response to dict.
        
        Args:
            response (object): Response object as returned by requests
        
        Returns:
            dict: A dict representing the JSON input or None in error case
        """
        try:
            return response.json()
        except Exception:
            raise DatabricksApiException(403, 3, "Invalid json message: %s" % self._remove_tags(response.text))

    def get(self, url, api_version, path, token):
        """REST GET function
        
        Args:
            url (str): URL for the request
            api_version (str): "1.2" (command execution) or "2.0" for all other API calls
            path (str): Additional path for the request
            token (str): Authentication token
        
        Returns:
            dict: A dict representing the JSON result of the call 
                  or in error case with keys status_code, error_code, message.
        """
        full_url = os.path.join(url, "api/%s" % api_version, path)
        try:
            response = requests.get(full_url, auth=("token", token))
        except Exception as ex:
            raise DatabricksApiException(0, 5, str(ex))
        if response.status_code in (200, 201):
            return self._json(response)
        else:
            raise DatabricksApiException(response.status_code, 1, self._remove_tags(response.text))

    def post(self, url, api_version, path, token, json=None, data=None, files=None):
        """REST POST function
        
        Args:
            url (str): URL for the request
            api_version (str): "1.2" (command execution) or "2.0" for all other API calls
            path (str): Additional path for the request
            token (str): Authentication token
            json (str, optional): json data to send in the body of the request. Defaults to None.
            data ([type], optional): Dictionary, list of tuples, bytes, or file-like object to send in 
                                     the body of the request. Defaults to None.
            files ([type], optional): Files to send with the request. Defaults to None.
            key (str, optional): Key to select from the converted JSON string. Defaults to None.
        
        Returns:
            dict: A dict representing the JSON result of the call 
            or in error case with keys status_code, error_code, message.
        """
        full_url = os.path.join(url, "api/%s" % api_version, path)
        try:
            response = requests.post(full_url, json=json, data=data, files=files, auth=("token", token))
        except Exception as ex:
            raise DatabricksApiException(0, 5, str(ex))
        if response.status_code in (200, 201):
            return self._json(response)
        else:
            raise DatabricksApiException(response.status_code, 2, self._remove_tags(response.text))


class Context(Rest):
    """Execution Context for Databricks

    Context is derived from ``Rest`` and uses Databricks REST API 1.2 
    
    Args:
        url (str): Databricks workspace URL
        cluster_id (str): Cluster ID
        token (str): Authentication token

    Attributes:
        url (str): Databricks workspace URL
        cluster_id (str): Cluster ID
        token (str): Authentication token
        id (str): Context ID
    """
    def __init__(self, url, cluster_id, token):
        self.token = token
        self.url = url
        self.cluster_id = cluster_id
        self.id = None

    def create(self):
        """Create an Execution Context
        
        Returns:
            str: Context ID
        """
        data = {"language": "python", "clusterId": self.cluster_id}
        response = self.post(self.url, "1.2", "contexts/create", data=data, token=self.token)
        self.id = response.get("id", None)
        if self.id is None:
            raise DatabricksApiException(403, 4, "Context ID missing")

    def status(self):
        """Get status of Execution Context
        
        Returns:
            dict: Dictionary with context id and status
        """
        path = "contexts/status?contextId=%s&clusterId=%s" % (self.id, self.cluster_id)
        return self.get(self.url, "1.2", path, token=self.token)

    def destroy(self):
        """Destroy Execution Context
        
        Returns:
            dict: Dictionary with context id
        """
        data = {"contextId": self.id, "clusterId": self.cluster_id}
        return self.post(self.url, "1.2", "contexts/destroy", data=data, token=self.token)


class Command(Rest):
    """Command execution on Databricks

    Command is derived from ``Rest`` and uses Databricks REST API 1.2 
        
    Args:
        url (str): Databricks workspace URL
        clusterId (str): Cluster ID
        token (str): Authentication token

    Attributes:
        url (str): Databricks workspace URL
        cluster_id (str): Cluster ID
        token (str): Authentication token
        context (Context): The Execution Context for the commands
    """
    def __init__(self, url, cluster_id, token):
        self.token = token
        self.url = url
        self.cluster_id = cluster_id
        self.context = Context(url, cluster_id, token)
        self.context.create()

    def execute(self, command):
        """Execute a python command under the given Execution Context
        
        Args:
            command (str): One line of python code
        
        Returns:
            str: The ID of the command
        """
        data = {"language": "python", "contextId": self.context.id, "clusterId": self.cluster_id}
        files = {"command": command}
        response = self.post(self.url, "1.2", "commands/execute", data=data, files=files, token=self.token)
        command_id = response.get("id", None)
        if command_id is None:
            raise DatabricksApiException(403, 4, "Command ID missing")
        
        finished = False
        while not finished:
            print(".", end="", flush=True)
            time.sleep(1)
            result = self.status(command_id)
            finished = result["status"] == "Finished"

        if result["results"]["resultType"] == "error":
            return (-1, result["results"]["cause"])
        else:
            return (0, result["results"]["data"])

    def status(self, command_id):
        """Check the status of the command execution
        
        Args:
            commandId (str): The command ID
        
        Returns:
            dict: Dictionary with context id and status
        """
        path = "commands/status?commandId=%s&contextId=%s&clusterId=%s" % (command_id, self.context.id, self.cluster_id)
        return self.get(self.url, "1.2", path, token=self.token)

    def close(self):
        self.context.destroy()
        

class Clusters(Rest):
    """Databricks Clusters API

    Clusters is derived from ``Rest`` and uses Databricks REST API 2.0
        
    Args:
        url (str): Databricks workspace URL
        token (str): Authentication token

    Attributes:
        url (str): Databricks workspace URL
        token (str): Authentication token
    """
    def __init__(self, url, token):
        self.url = url
        self.token = token

    def status(self, cluster_id):
        """Status of cluster
        
        Args:
            cluster_id (str): Cluster ID
        
        Returns:
            dict:  Cluster Response Dictionary (https://docs.databricks.com/api/latest/clusters.html#clustergetclusterresponse)
        """
        return self.get(self.url, "2.0", "clusters/get?cluster_id=%s" % cluster_id, token=self.token)

    def start(self, cluster_id):
        """Start a cluster on Databricks
        
        Args:
            cluster_id (str): Cluster ID of cluster to start
        
        Returns:
            dict:  Cluster Response Dictionary (https://docs.databricks.com/api/latest/clusters.html#clustergetclusterresponse)
        """
        data = {"cluster_id": cluster_id}
        return self.post(self.url, "2.0", "clusters/start", json=data, token=self.token)

    def edit(self, data):
        """Start a cluster on Databricks
        
        Args:
            cluster_id (str): Cluster ID of cluster to start
        
        Returns:
            dict:  Cluster Response Dictionary (https://docs.databricks.com/api/latest/clusters.html#clustergetclusterresponse)
        """
        return self.post(self.url, "2.0", "clusters/edit", json=data, token=self.token)


class Libraries(Rest):
    """Databricks Library API

    Libraries is derived from ``Rest`` and uses Databricks REST API 2.0
        
    Args:
        url (str): Databricks workspace URL
        token (str): Authentication token

    Attributes:
        url (str): Databricks workspace URL
        token (str): Authentication token
    """
    def __init__(self, url, token):
        self.url = url
        self.token = token

    def status(self, cluster_id):
        """Status of library installation
        
        Args:
            cluster_id (str): Cluster ID
        
        Returns:
            dict:  Library Response Dictionary (https://docs.databricks.com/api/latest/clusters.html#clustergetclusterresponse)
        """
        return self.get(self.url, "2.0", "libraries/cluster-status?cluster_id=%s" % cluster_id, token=self.token)



import os
from databrickslabs_jupyterlab._version import __version__, __version_info__

if os.environ.get("DBJL_HOST") is None:

    from databrickslabs_jupyterlab.status import KernelHandler, DbStartHandler, DbStatusHandler, Status
    from notebook.utils import url_path_join

    def load_jupyter_server_extension(nbapp):
        """
        Called during notebook start
        """
        print("Loading server extension ...")
        KernelHandler.nbapp = nbapp
        base_url = nbapp.web_app.settings['base_url']
        status_route = url_path_join(base_url, '/databrickslabs-jupyterlab-status')
        start_route = url_path_join(base_url, '/databrickslabs-jupyterlab-start')
        nbapp.web_app.add_handlers('.*', [(status_route, DbStatusHandler), (start_route, DbStartHandler)])



import json
import os
import re
import subprocess
import sys
import tempfile

import databrickslabs_jupyterlab
from databrickslabs_jupyterlab.remote import get_python_path, get_remote_packages
from databrickslabs_jupyterlab.local import print_ok

WHITELIST = [
    "hyperopt", "keras-applications", "keras-preprocessing", "keras", "matplotlib", "mleap", "mlflow", "numba", "numpy",
    "pandas", "patsy", "pillow", "pyarrow", "python-dateutil", "pyparsing", "scikit-learn", "scipy", "seaborn",
    "simplejson", "statsmodels", "tabulate", "tensorboard", "tensorboardx", "tensorflow-estimator", "tensorflow",
    "torch", "torchvision"
]

BLACKLIST = [
    "absl-py", "ansi2html", "asn1crypto", "astor", "attrs", "backcall", "backports.shutil-get-terminal-size", "bcrypt",
    "bleach", "blessings", "boto", "boto3", "botocore", "brewer2mpl", "certifi", "cffi", "chardet", "click", "colorama",
    "configobj", "configparser", "cryptography", "cycler", "cython", "databricks-cli", "databrickslabs-jupyterlab",
    "decorator", "defusedxml", "docker", "docopt", "docutils", "entrypoints", "enum34", "et-xmlfile", "flask",
    "freetype-py", "funcsigs", "fusepy", "future", "gast", "gitdb2", "gitpython", "grpcio", "gunicorn", "h5py",
    "horovod", "html5lib", "idna", "inquirer", "ipaddress", "ipykernel", "ipython", "ipython-genutils", "ipywidgets",
    "itsdangerous", "jdcal", "jedi", "jinja2", "jmespath", "jsonschema", "jupyter-client", "jupyter-core", "llvmlite",
    "lxml", "mako", "markdown", "markupsafe", "mistune", "mkl-fft", "mkl-random", "mock", "msgpack", "msgpack-python",
    "nbconvert", "nbformat", "ndg-httpsclient", "networkx", "nose", "nose-exclude", "notebook", "olefile", "openpyxl",
    "openpyxl", "packaging", "pandocfilters", "paramiko", "parso", "pathlib2", "pexpect", "pickleshare", "pip", "ply",
    "prometheus-client", "prompt-toolkit", "protobuf", "psutil", "psycopg2", "ptyprocess", "py4j", "pyasn1",
    "pycparser", "pycurl", "pygments", "pygobject", "pymongo", "pynacl", "pypng", "pysocks", "python-apt",
    "python-dateutil", "python-editor", "python-geohash", "pyopenssl", "pytz", "pyyaml", "pyzmq", "querystring-parser",
    "readchar", "requests", "scour", "send2trash", "setuptools", "sidecar", "simplegeneric", "simplejson",
    "singledispatch", "six", "smmap2", "sqlparse", "ssh-config", "ssh-import-id", "tabulate", "termcolor", "terminado",
    "testpath", "texttable", "tornado", "tqdm", "traitlets", "unattended-upgrades", "urllib3", "virtualenv", "wcwidth",
    "webencodings", "websocket-client", "werkzeug", "wheel", "widgetsnbextension", "wrapt"
]


def execute(script, script_name, message):
    with tempfile.TemporaryDirectory() as tmpdir:
        script_file = os.path.join(tmpdir, script_name)
        with open(script_file, "w") as fd:
            fd.write(script)

        try:
            result = subprocess.call(["bash", script_file])
            if result != 0:
                sys.exit(1)
        except Exception as ex:
            print(message, ex)
            sys.exit(1)


def update_env(env_file):
    script = """#!/bin/bash
conda env update --file %s
""" % (env_file)
    print_ok("Updating current conda environment")
    execute(script, "update_env.sh", "Error while updating conda environment")


def install_env(env_file, env_name=None):
    script = """#!/bin/bash
conda env create -n %s -f %s
source $(conda info | awk '/base env/ {print $4}')/bin/activate "%s" 
""" % (env_name, env_file, env_name)

    print_ok("   => Installing conda environment %s" % env_name)
    execute(script, "install_env.sh", "Error while installing conda environment")


def install_labextensions(labext_file, env_name=None):
    if env_name is None:
        script = """#!/bin/bash
"""        
    else:
        script = """#!/bin/bash
source $(conda info | awk '/base env/ {print $4}')/bin/activate "%s" 
""" % env_name

    script += ("""
jupyter labextension install $(cat %s)
jupyter lab build
""" % labext_file)

    print_ok("   => Installing jupyterlab extensions")
    execute(script, "install_labext.sh", "Error while installing jupyter labextensions")


def update_local():
    module_path = os.path.dirname(databrickslabs_jupyterlab.__file__)

    env_file = os.path.join(module_path, "lib/env.yml")
    update_env(env_file)

    labext_file = os.path.join(module_path, "lib/labextensions.txt")
    install_labextensions(labext_file)
    usage(os.environ.get("CONDA_DEFAULT_ENV", "unknown"))


def install(profile, host, token, cluster_id, cluster_name, use_whitelist):
    print("\n* Installation of local environment to mirror a remote Databricks cluster")
    libs = get_remote_packages(cluster_id, host, token)

    if use_whitelist:
        print_ok("   => Using whitelist to select packages")
        ds_libs = [lib for lib in libs if lib["name"].lower() in WHITELIST]
    else:
        print_ok("   => Using blacklist to select packages")
        ds_libs = [lib for lib in libs if lib["name"].lower() not in BLACKLIST]

    ds_yml = ""
    for lib in ds_libs:
        if lib["name"] == "hyperopt":
            r = re.compile(r"(\d+\.\d+.\d+)(.*)")
            version = r.match(lib["version"]).groups()[0]
        else:
            version = lib["version"]
        ds_yml += ("    - %s==%s\n" % (lib["name"], version))
    print("\n    Library versions being installed:")
    print(ds_yml + "\n")

    module_path = os.path.dirname(databrickslabs_jupyterlab.__file__)
    env_file = os.path.join(module_path, "lib/env.yml")
    with open(env_file, "r") as fd:
        master_yml = fd.read()

    with tempfile.TemporaryDirectory() as tmpdir:
        env_file = os.path.join(tmpdir, "env.yml")
        with open(env_file, "w") as fd:
            fd.write(master_yml)
            fd.write("\n    # Data Science Libs\n")
            fd.write(ds_yml)
            fd.write("\n")

        env_name = cluster_name.replace(" ", "_")
        answer = input("    => Provide a conda environment name (default = %s): " % env_name)
        if answer != "":
            env_name = answer.replace(" ", "_")

        install_env(env_file, env_name)

        labext_file = os.path.join(module_path, "lib/labextensions.txt")
        install_labextensions(labext_file, env_name)

    usage(env_name)


def usage(env_name):
    print_ok("""
= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
                              Q U I C K S T A R T
= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 

Prerequisites:
--------------

- Install Databricks CLI and configure profile(s) for your cluster(s)

  AWS: https://docs.databricks.com/user-guide/dev-tools/databricks-cli.html)
  Azure: https://docs.azuredatabricks.net/user-guide/dev-tools/databricks-cli.html

- Create an ssh key pair called ~/.ssh/id_<profile> for each cluster and 
  add the public key to the cluster SSH configuration


databrickslabs-jupyterlab:
--------------------------

1) Show help

    conda activate %s
    databrickslabs-jupyterlab -h

2) Create jupyter kernel for remote cluster

    Databricks on AWS:
        databrickslabs-jupyterlab <profile> -k [-i cluster-id]
    
    Azure Databricks:
        databrickslabs-jupyterlab <profile> -k -o <organisation>

3) Compare local and remote python package versions
    
    databrickslabs-jupyterlab <profile> -v all|same|diff

4) Copy Personal Access token for databricks cluster to cipboard (same on AWS and Azure)

    databrickslabs-jupyterlab <profile> -c

5) Start jupyter lab to use the kernel(s) created in 2)

    databrickslabs-jupyterlab <profile> -l [-i cluster-id]


6) Check currently available profiles 

    databrickslabs-jupyterlab -p

""" % env_name)



import threading
import time
import uuid

from IPython.display import HTML, display
from IPython import get_ipython

try:
    from ipywidgets import FloatProgress, Label, HBox, VBox, Layout, Button
    has_ipywidgets = True
except:
    has_ipywidgets = False


class Progress(object):
    def __init__(self, sc, job_info):
        self.sc = sc
        self.job_info = job_info
        self.tracker = sc.statusTracker()
        self.progressbar_showing = False

    def worker(self):
        def cancel(b):
            self.sc.cancelJobGroup(self.job_info.group_id)

        def toggle(widget):
            def f(b):
                for w in widget.children:
                    h = w.layout.height
                    if h is None or h == "16px":
                        w.layout.height = "0px"
                        b.icon = 'arrow-circle-down'
                    else:
                        w.layout.height = "16px"
                        b.icon = 'arrow-circle-right'

            return f

        style = {'description_width': 'initial'}
        bars = {}
        labels = {}
        lastJob = None

        progressbars = VBox([])

        cancel_button = Button(button_style='', tooltip='Cancel Spark Job', icon='window-close')
        cancel_button.add_class("db-button")
        cancel_button.on_click(cancel)

        toggle_button = Button(button_style='', tooltip='Toggle progress bar', icon='arrow-circle-right')
        toggle_button.add_class("db-button")
        toggle_button.on_click(toggle(progressbars))

        indicator = HBox([toggle_button, progressbars])

        # print(self.running, self.progressbar_showing)

        while (self.running == 1):
            time.sleep(0.1)
            jobs = [(jobid, self.tracker.getJobInfo(jobid))
                    for jobid in self.tracker.getJobIdsForGroup(self.job_info.group_id)
                    if self.tracker.getJobInfo(jobid).status == "RUNNING"]

            for j, job in jobs:
                if bars.get(j, None) is None:
                    if lastJob is not None:
                        bars[lastJob].value = 100.0
                    bars[j] = FloatProgress(value=0.0,
                                            min=0.0,
                                            max=100.0,
                                            description='Job: % 3d Stage: % 3d' % (j, 0),
                                            bar_style='info',
                                            orientation='horizontal',
                                            style=style)
                    bars[j].add_class("db-bar")
                    labels[j] = Label(value='',
                                      description='Code:',
                                      disabled=False,
                                      layout=Layout(width="800px", height="100%", margin="0 0 0 5px"))
                    labels[j].add_class("db-label")

                    progressbar = HBox([bars[j], labels[j]])
                    progressbars.children = progressbars.children + (progressbar, )
                    if not self.progressbar_showing:
                        self.progressbar_showing = True
                        display(indicator)

                lastJob = j
                stageIds = sorted(job.stageIds)
                for s in stageIds:
                    stageInfo = self.tracker.getStageInfo(s)
                    if stageInfo.numActiveTasks > 0:
                        bars[j].description = 'Job: %04d Stage: %04d' % (j, s)
                        progress = int(100 * stageInfo.numCompletedTasks / stageInfo.numTasks)
                        bars[j].value = progress
                        labels[j].value = "code: '%s' / stages: %s" % (stageInfo.name, str(stageIds)[1:-1])

        if lastJob is not None and self.running == 0:
            bars[lastJob].value = 100.0

    def start_progressbar(self, info):
        # print("start_progressbar")
        self.sc.setLocalProperty("spark.scheduler.pool", self.job_info.pool_id)
        self.job_info.group_id = self.job_info.pool_id + "_" + uuid.uuid4().hex
        self.sc.setJobGroup(self.job_info.group_id, "jupyterlab job group", True)

        time.sleep(0.1)

        self.running = 1
        self.progressbar_showing = False
        self.t = threading.Thread(target=self.worker)
        self.t.start()

    def stop_progressbar(self, result):
        # print("stop_progressbar")
        def is_running():
            job_status = [
                self.tracker.getJobInfo(jobid).status
                for jobid in self.tracker.getJobIdsForGroup(self.job_info.group_id)
            ]
            return any(status == "RUNNING" for status in job_status)

        if is_running():
            print("Killing spark job ", end="")
            while is_running():
                self.sc.cancelJobGroup(self.job_info.group_id)
                print(".", end="")
                time.sleep(1)
            print(". Spark job killed")
            self.running = -1
        else:
            # print("done")
            self.running = 0


def load_css():
    if has_ipywidgets:
        display(
            HTML("""
        <style>
        .db-bar {
            height: 12px;
        }

        .db-label {
            height: 14px;
            line-height: 11px !important;
            font-size: 11px;
            color: #999 !important;
        }
        div.db-bar label {
            font-size: 11px;
            color: #999 !important;
        }
        .db-button {
            width: 13px;
            height: 12px;
            padding: 0px;
        }
        .db-button .fa {
            display: block;
        }
        </style>
        """))


def load_progressbar(ip, sc, job_info):
    if has_ipywidgets:
        progress = Progress(sc, job_info)
        for event, name, func in (('pre_run_cell', "start_progressbar", progress.start_progressbar),
                                  ('post_run_cell', "stop_progressbar", progress.stop_progressbar)):
            register = True
            for m in ip.events.callbacks["pre_run_cell"]:
                if m.__name__ == name:
                    register = False
            if register:
                ip.events.register(event, func)



from collections import namedtuple
import re

VersionInfo = namedtuple('VersionInfo', ['major', 'minor', 'patch', 'release', 'build'])


def get_version(version):
    r = re.compile(r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)\-{0,1}(?P<release>\D*)(?P<build>\d*)')
    major, minor, patch, release, build = r.match(version).groups()
    return VersionInfo(major, minor, patch, release, build)


__version__ = "1.0.2-rc2"  # DO NOT EDIT THIS DIRECTLY!  It is managed by bumpversion
__version_info__ = get_version(__version__)



import json
import os
import sys
import time
from os.path import expanduser
import socket
import subprocess
import uuid
import glob

import inquirer

from databricks_cli.configure.provider import get_config, ProfileConfigProvider
from databricks_cli.sdk.api_client import ApiClient
from databricks_cli.clusters.api import ClusterApi

import databrickslabs_jupyterlab
from databrickslabs_jupyterlab._version import __version__
from databrickslabs_jupyterlab.rest import Command, Clusters, Libraries, DatabricksApiException
from databrickslabs_jupyterlab.utils import (bye, Dark, print_ok, print_error, print_warning, run)


def connect(profile):
    """Initialize Databricks API client
    
    Args:
        profile (str): Databricks CLI profile string
    
    Returns:
        ApiClient: Databricks ApiClient object
    """
    config = ProfileConfigProvider(profile).get_config()
    if config is None:
        print_error("Cannot initialize ApiClient")
        bye(1)
    verify = config.insecure is None
    if config.is_valid_with_token:
        return ApiClient(host=config.host, token=config.token, verify=verify)
    else:
        print_error("Token for profile '%s' is invalid" % profile)
        bye(1)


def select_cluster(clusters):
    """Build a list of clusters for the user to select one
    
    Args:
        clusters (list): list of clusters
    
    Returns:
        dict: Cluster config of selected cluster
    """
    def entry(i, cluster):
        if cluster.get("autoscale", None) is None:
            return "%s: '%s' (id: %s, state: %s, workers: %d)" % (i, cluster["cluster_name"], cluster["cluster_id"],
                                                                cluster["state"], cluster["num_workers"])
        else:
            return "%s: '%s' (id: %s, state: %s, scale: %d-%d)" % (i, cluster["cluster_name"], cluster["cluster_id"],
                                                                 cluster["state"], cluster["autoscale"]["min_workers"],
                                                                 cluster["autoscale"]["max_workers"])

    choice = [
        inquirer.List('cluster_id',
                      message='Which cluster to connect to?',
                      choices=[entry(i, cluster) for i, cluster in enumerate(clusters)])
    ]
    answer = inquirer.prompt(choice, theme=Dark())
    return clusters[int(answer["cluster_id"].split(":")[0])]


def get_cluster(profile, host, token, cluster_id=None, status=None):
    """Get the cluster configuration from remote
    
    Args:
        profile (str): Databricks CLI profile string
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile stringf
        cluster_id (str, optional): If cluster_id is given, the user will not be asked to select one. Defaults to None.
        status (Status, optional): A Status class providing set_status. Defaults to None.
    
    Returns:
        tuple: Cluster configs: cluster_id, public_ip, cluster_name and a flag whether it was started
    
    Returns:
        tuple: (cluster_id, public_ip, cluster_name, started)
    """
    with open("%s/.ssh/id_%s.pub" % (expanduser("~"), profile)) as fd:
        try:
            ssh_pub = fd.read().strip()
        except:
            print_error("   Error: ssh key for profile 'id_%s.pub' does not exist in %s/.ssh" % (profile, expanduser("~")))
            bye()

    try:
        apiclient = connect(profile)
        client = ClusterApi(apiclient)
        clusters = client.list_clusters()
    except Exception as ex:
        print_error(ex)
        return (None, None, None, None)

    clusters = clusters["clusters"]

    if cluster_id is not None:
        cluster = None
        for c in clusters:
            if c["cluster_id"] == cluster_id:
                cluster = c
                break

        if cluster is None:
            print_error("   Error: A cluster with id '%s' does not exist in the workspace of profile '%s'" %
                        (cluster_id, profile))
            return (None, None, None, None)

        if ssh_pub not in [c.strip() for c in cluster.get("ssh_public_keys", [])]:
            print_error("   Error: Cluster with id '%s' does not have ssh key '~/.ssh/id_%s' configured" %
                        (cluster_id, profile))
            return (None, None, None, None)
    else:
        my_clusters = [
            cluster for cluster in clusters
            if ssh_pub in [c.strip() for c in cluster.get("ssh_public_keys", [])]
        ]

        if not my_clusters:
            print_error(
                "    Error: There is no cluster in the workspace for profile '%s' configured with ssh key '~/.ssh/id_%s':" %
                (profile, profile))
            print("    Use 'databrickslabs_jupyterlab %s -s' to configure ssh for clusters in this workspace\n" % profile)
            return (None, None, None, None)

        current_conda_env = os.environ.get("CONDA_DEFAULT_ENV", None)
        found = None
        for i, c in enumerate(my_clusters):
            if c["cluster_name"].replace(" ", "_") == current_conda_env:
                found = c["cluster_name"]
                break

        if found is not None:
            print_warning(
                "\n   => The current conda environment is '%s'.\n      You might want to select cluster %d with the name '%s'?\n" %
                (current_conda_env, i, found))

        cluster = select_cluster(my_clusters)

    cluster_id = cluster["cluster_id"]
    cluster_name = cluster["cluster_name"]

    cluster_api = Clusters(url=host, token=token)
    try:
        response = cluster_api.status(cluster_id)
    except DatabricksApiException as ex:
        print_error(ex)
        return (None, None, None, None)

    state = response["state"]
    if not state in ["RUNNING", "RESIZING"]:
        if state == "TERMINATED":
            print("   => Starting cluster %s" % cluster_id)
            if status is not None:
                status.set_status(profile, cluster_id, "Cluster Starting")
            try:
                response = cluster_api.start(cluster_id)
            except DatabricksApiException as ex:
                print_error(ex)
                return (None, None, None, None)

        print("   => Waiting for cluster %s being started (this can take up to 5 min)" % cluster_id)
        print("   ", end="", flush=True)

        while not state in ("RUNNING", "RESIZING"):
            if status is not None:
                status.set_status(profile, cluster_id, "Cluster Starting", False)
            print(".", end="", flush=True)
            time.sleep(5)
            try:
                response = cluster_api.status(cluster_id)
            except DatabricksApiException as ex:
                print_error(ex)
                return (None, None, None, None)

            if response.get("error", None) is not None:
                print("ERROR", response)
                state = "UNKNOWN"
            else:
                state = response["state"]
        print_ok("\n   => OK")

        if status is not None:
            status.set_status(profile, cluster_id, "Cluster started")

        print("\n   => Waiting for libraries on cluster %s being installed (this can take some time)" % cluster_id)
        print("   ", end="", flush=True)

        done = False
        while not done:
            try:
                states = get_library_state(cluster_id, host=host, token=token)
            except DatabricksApiException as ex:
                print_error(ex)
                return (None, None, None, None)
            installing = any([s in ["PENDING", "RESOLVING", "INSTALLING"] for s in states])
            if installing:
                if status is not None:
                    status.set_status(profile, cluster_id, "Installing cluster libraries", False)
                print(".", end="", flush=True)
                time.sleep(5)
            else:
                done = True
                print_ok("\n   => OK\n")

        if status is not None:
            status.set_status(profile, cluster_id, "Cluster libraries installed", False)

    public_ip = response["driver"].get("public_dns", None)
    if public_ip is None:
        print_error("   Error: Cluster does not have public DNS name")
        return (None, None, None, None)

    print_ok("   => Selected cluster: %s (%s)" % (cluster_name, public_ip))

    return (cluster_id, public_ip, cluster_name, None)


def get_python_path(host, conda_env=None):
    # conda_default_env = ssh(host, "echo $DEFAULT_DATABRICKS_ROOT_CONDA_ENV").strip().decode("utf-8")
    # if conda_default_env == "":
    #     python_path = "/databricks/python3/bin"
    #     print_ok("   => %s (pip managed cluster)" % python_path)
    # else:
    #     if conda_env is None:  # if there is no conda environment variable, use default
    #         conda_env = conda_default_env
    #     python_path = "/databricks/conda/envs/%s/bin" % conda_env
    #     print_ok("   => %s (conda managed cluster)" % python_path)

    python_path = "/databricks/python/bin"
    return python_path


def install_libs(cluster_id, host, token):
    """Install ipywidgets, sidecar and databrickslabs_jupyterlab libraries on the driver
    
    Args:
        host (str): host from databricks cli config for given profile string
        python_path (str): Remote python path to be used for kernel
    """
    python_path = get_python_path(cluster_id)

    packages = json.loads(subprocess.check_output(["conda", "list", "--json"]))
    deps = {p["name"]: p["version"] for p in packages if p["name"] in ["ipywidgets", "sidecar"]}
    libs = ["ipywidgets==%s" % deps["ipywidgets"],
            "sidecar==%s" % deps["sidecar"],
            "databrickslabs-jupyterlab==%s" % __version__]
    pip_cmd = ["%s/pip" %python_path,  "install", "-q", "--no-warn-conflicts", "--disable-pip-version-check"] + libs

    cmd =  'import subprocess, json; ' + \
          ('ret = subprocess.run(%s, stderr=subprocess.PIPE, encoding="utf-8"); ' % pip_cmd) + \
           'print(json.dumps([ret.returncode, str(ret.stderr)]))'

    print("   => Installing %s" % ", ".join(libs))
    try:
        command = Command(url=host, cluster_id=cluster_id, token=token)
        result = command.execute(cmd)
        command.close()
    except DatabricksApiException as ex:
        print("REST API error", ex)
        return False

    if result[0] == 0:
        return tuple(json.loads(result[1]))
    else:
        return result


def get_remote_packages(cluster_id, host, token):
    cmd = 'import pkg_resources, json; print(json.dumps([{"name":p.key, "version":p.version} for p in pkg_resources.working_set]))'
    try:
        command = Command(url=host, cluster_id=cluster_id, token=token)
        result = command.execute(cmd)
        command.close()
    except DatabricksApiException as ex:
        print(ex)
        return None
    return result

def is_reachable(public_dns):
    """Check whether a remote cluster is reachable
    
    Args:
        public_dns (str): Public IP address or DNS name
    
    Returns:
        bool: True if reachable else False
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(3)
    result = sock.connect_ex((public_dns, 2200))
    try:
        sock.close()
    except:
        pass
    return result == 0


def get_library_state(cluster_id, host, token):
    """Get the state of the library installation on the remote cluster
    
    Args:
        cluster_id (str): Cluster ID
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile stringf
    
    Returns:
        list: list of installation status for each custom library
    """
    libraries_api = Libraries(url=host, token=token)
    libraries = libraries_api.status(cluster_id)

    if libraries.get("library_statuses", None) is None:
        return []
    else:
        return [lib["status"] for lib in libraries["library_statuses"]]


def check_installed(cluster_id, host, token):
    """Check whether databrickslabs_jupyterlab is installed on the remote host
    
    Args:
    
    Returns:
        bool: True if installed else False
    """
    result = get_remote_packages(cluster_id, host, token)
    if result[0] == 0:
        packages = json.loads(result[1])
        package_names = [p["name"] for p in packages]
        needed = ["databrickslabs-jupyterlab", "ipywidgets", "ipykernel"]

        return all([p in package_names for p in needed])
    else:
        print(result[1])
        return False


def version_check(cluster_id, host, token, flag):
    """Compare local and remote library versions
    
    Args:
        cluster_id (str): Cluster ID
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile string
        flag (str): all|diff|same
    """
    def normalize(key):
        return key.lower().replace("-", "_")

    packages = json.loads(subprocess.check_output(["conda", "list", "--json"]))
    deps = {normalize(p["name"]): p["version"] for p in packages}

    result = get_remote_packages(cluster_id, host, token)
    if result[0] == 0:
        remote_packages = json.loads(result[1])
    else:
        return
    remote_deps = {normalize(p["name"]): p["version"] for p in remote_packages}
    joint_keys = sorted(list(set(list(deps.keys()) + list(remote_deps.keys()))))
    print("%-30s %-10s%-10s" % ("Package", "local", "remote"))
    if str(flag) == "all":
        scope = joint_keys
    elif str(flag) == "same":
        scope = [key for key in joint_keys if deps.get(key, None) == remote_deps.get(key, None)]
    else:
        scope = [
            key for key in joint_keys if deps.get(key, None) != remote_deps.get(key, None)
            #            and deps.get(key, None) is not None and remote_deps.get(key, None) is not None
        ]
    for key in scope:
        result = "%-30s %-10s  %-10s" % (key, deps.get(key, "--"), remote_deps.get(key, "--"))
        if deps.get(key) == remote_deps.get(key):
            print_ok(result)
        else:
            print_error(result)

def configure_ssh(profile, host, token, cluster_id):
    """Configure SSH for the remote cluster
    
    Args:
        profile (str): Databricks CLI profile string
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile string
        cluster_id (str): cluster ID
    """
    sshkey_file = os.path.expanduser("~/.ssh/id_%s" % profile)
    if not os.path.exists(sshkey_file):
        print("\n   => ssh key '%s' does not exist" % sshkey_file)
        answer = input("   => Shall it be created (y/n)? (default = n): ")
        if answer.lower() == "y":
            print("   => Creating ssh key %s" % sshkey_file)
            subprocess.call(["ssh-keygen", "-b", "2048",  "-N", "", "-f",  sshkey_file])
            print_ok("   => OK")
        else:
            bye()
    else:
        print_ok("\n   => ssh key '%s' already exists" % sshkey_file)

    with open(sshkey_file + ".pub", "r") as fd:
        sshkey = fd.read().strip()

    cluster_api = Clusters(url=host, token=token)
    try:
        response = cluster_api.status(cluster_id)
    except DatabricksApiException as ex:
        print_error(ex)
        return None

    ssh_public_keys = response.get("ssh_public_keys", [])
    if sshkey in [key.strip() for key in ssh_public_keys]:
        print_ok("   => public ssh key already configured for cluster %s" % cluster_id)
        bye()

    request = {}
    for key in ["cluster_id", "cluster_name", "spark_version", "node_type_id"]:
        request[key] = response[key]

    if response.get("num_workers", None) is not None:
        request["num_workers"] = response["num_workers"]

    if response.get("autoscale", None) is not None:
        request["autoscale"] = response["autoscale"]

    request["ssh_public_keys"] = ssh_public_keys + [sshkey]

    print_warning("   => The ssh key will be added to the cluster. \n   Note: The cluster will be restarted immediately!")
    answer = input("   => Shall the ssh key be added and the cluster be restarted (y/n)? (default = n): ")
    if answer.lower() == "y":
        try:
            response = cluster_api.edit(request)
        except DatabricksApiException as ex:
            print_error(str(ex))
            return None
        print_ok("   => OK")
    else:
        print_error("   => Cancelled")



import os
from collections import defaultdict
from pexpect import pxssh
import random
import ssh_config
import subprocess
import sys
import time

from inquirer.themes import Default, term


# class Ssh:
#     def __init__(self, ssh_config="~/.ssh/config"):
#         self.ssh_config = os.path.expanduser(ssh_config)
#         self.ssh = defaultdict(lambda: None)

#     def _split_lines(self, result):
#         return result.decode("utf-8").strip().split("\r\n")[1:]

#     def execute(self, host, cmd):
#         if self.ssh[host] is None:
#             success = False
#             delays = [2, 4, 8]
#             for retry in range(3):
#                 print("   => Connecting to %s via ssh (%d)" % (host, retry + 1))
#                 try:
#                     self.ssh[host] = pxssh.pxssh(options={"StrictHostKeyChecking": "no"})
#                     self.ssh[host].login(host, quiet=False, ssh_config=self.ssh_config)
#                     success = True
#                     break
#                 except:
#                     delay = round(random.random(), 2) + delays[retry]
#                     print_warning("   Failed to log into %s, waiting %3.1f seconds" % (host, delay))
#                     time.sleep(delay)

#             if success:
#                 print_ok("   => OK")
#             else:
#                 print_error("   => Error")
#                 return {"exit_code": -1, "result": None}

#         self.ssh[host].sendline(cmd)
#         self.ssh[host].prompt()
#         result = self._split_lines(self.ssh[host].before)
#         self.ssh[host].sendline('echo $?')
#         self.ssh[host].prompt()
#         exit_code = self._split_lines(self.ssh[host].before)
#         try:
#             exit_code = int(exit_code[0])
#         except:
#             pass
#         if exit_code != 0: 
#             print_error("   => Remote command failed")
#         return {"exit_code": exit_code, "result": result}

#     def close(self, host):
#         if self.ssh[host] is not None:
#             print("   => Closing ssh connection to %s" % host)
#             self.ssh[host].logout()
#             self.ssh[host] = None


def run(cmd):
    """Run a shell command
    
    Args:
        cmd (str): shell command
    """
    try:
        subprocess.run(cmd)
    except:
        print_error("Error running: %s" % cmd)
        bye(1)


def bye(status=0):
    """Standard exit function
    
    Args:
        msg (str, optional): Exit message to be printed. Defaults to None.
    """
    if status != 0:
        print_error("\n=> Exiting")
    sys.exit(status)


class Dark(Default):
    """Dark Theme for inquirer"""
    def __init__(self):
        super().__init__()
        self.List.selection_color = term.cyan


class Colors:
    ERROR = '\033[91m'
    OK = '\033[92m'
    WARNING = '\033[93m'
    RESET = '\033[0m'


def _print(color, *args):
    print(color, end="")
    print(*args, end="")
    print(Colors.RESET)


def print_ok(*args):
    _print(Colors.OK, *args)


def print_error(*args):
    _print(Colors.ERROR, *args)


def print_warning(*args):
    _print(Colors.WARNING, *args)


class SshConfig:
    def __init__(self):
        self.config_file = os.path.expanduser("~/.ssh/config")
        self.load()

    def _get_mtime(self):
        return os.stat(self.config_file).st_mtime

    def load(self):
        self.mtime = self._get_mtime()
        sc = ssh_config.SSHConfig.load(self.config_file)
        self.hosts = {
            h.name: h.attributes()["HostName"]
            for h in sc.hosts() if h.attributes().get("HostName", None) is not None
        }

    def get_dns(self, host_name):
        if self._get_mtime() > self.mtime:
            self.load()
        return self.hosts.get(host_name, None)





import configparser
from jupyter_client import kernelspec
import os
import subprocess
import sys
import textwrap
import time
from os.path import expanduser
from ssh_config import SSHConfig, Host
import inquirer

from databrickslabs_jupyterlab.utils import (bye, Dark, print_ok, print_error, print_warning)


def conda_version():
    """Check conda version"""
    try:
        result = subprocess.check_output(["conda", "--version"])
    except Exception as ex:
        print(ex)
        print("conda cannot be called. Is it properly installed?")
        sys.exit(1)

    result = result.strip().decode()
    return tuple([int(v) for v in result.split(" ")[1].split(".")])


def write_config():
    """Store jupyter lab configuration necessary for databrickslabs_jupyterlab
    Set values:
    - c.KernelManager.autorestart: False
    - c.MappingKernelManager.kernel_info_timeout: 600
    """
    config = {"c.KernelManager.autorestart": False, "c.MappingKernelManager.kernel_info_timeout": 600}

    config_file = os.path.expanduser("~/.jupyter/jupyter_notebook_config.py")
    if os.path.exists(config_file):
        with open(config_file, "r") as fd:
            lines = fd.read().split("\n")

        with open(config_file, "w") as fd:
            for line in lines:
                kv = line.strip().split("=")
                if len(kv) == 2:
                    k, v = kv
                    if config.get(k, None) is not None:
                        fd.write("%s=%s\n" % (k, config[k]))
                        del config[k]
                    else:
                        fd.write("%s\n" % line)
            for k, v in config.items():
                fd.write("%s=%s\n" % (k, v))
    else:
        with open(config_file, "w") as fd:
            fd.write("\n".join(["%s=%s" % (k, v) for k, v in config.items()]))


def get_db_config(profile):
    """Get Databricks configuration from ~/.databricks.cfg for given profile
    
    Args:
        profile (str): Databricks CLI profile string
    
    Returns:
        tuple: The tuple of host and personal access token from ~/.databrickscfg
    """
    config = configparser.ConfigParser()
    configs = config.read(expanduser("~/.databrickscfg"))
    if not configs:
        print_error("Cannot read ~/.databrickscfg")
        bye(1)

    profiles = config.sections()
    if not profile in profiles:
        print(" The profile '%s' is not available in ~/.databrickscfg:" % profile)
        for p in profiles:
            print("- %s" % p)
        bye()
    else:
        host = config[profile]["host"]
        token = config[profile]["token"]
        return host, token


def prepare_ssh_config(cluster_id, profile, public_ip):
    """Add/edit the ssh configuration belonging to the given cluster in ~/.ssh/config
    
    Args:
        cluster_id (str): Cluster ID
        profile (str): Databricks CLI profile string
        public_ip (str): Public IP address
    """
    config = os.path.join(expanduser("~"), ".ssh/config")
    try:
        sc = SSHConfig.load(config)
    except:
        sc = SSHConfig(config)
    hosts = [h.name for h in sc.hosts()]
    if cluster_id in hosts:
        host = sc.get(cluster_id)
        host.set("HostName", public_ip)
        print("   => Added ssh config entry or modified IP address:\n")
        print(textwrap.indent(str(host), "      "))
    else:
        attrs = {
            'HostName': public_ip,
            'IdentityFile': '~/.ssh/id_%s' % profile,
            'Port': 2200,
            'User': 'ubuntu',
            'ServerAliveInterval': 300,
            'ServerAliveCountMax': 2
        }
        host = Host(name=cluster_id, attrs=attrs)
        print("Adding ssh config to ~/.ssh/config:\n")
        print(textwrap.indent(str(host), "      "))
        sc.append(host)
    sc.write()


def show_profiles():
    """Show locally configured Databricks CLI profile"""
    template = "%-20s %-60s %s"
    print("")
    print(template % ("PROFILE", "HOST", "SSH KEY"))
    config = configparser.ConfigParser()
    config.read(expanduser("~/.databrickscfg"))
    profiles = config.sections()

    for profile in profiles:
        host, _ = get_db_config(profile)
        ssh_ok = "OK" if os.path.exists(os.path.expanduser("~/.ssh/id_%s") % profile) else "MISSING"
        print(template % (profile, host, ssh_ok))


def create_kernelspec(profile, organisation, host, cluster_id, cluster_name, local_env, python_path):
    """Create or edit the remote_ikernel specification for jupyter lab
    
    Args:
        profile (str): Databricks CLI profile string    
        organisation (str): In case of Azure, the organization ID, else None
        host (str): host from databricks cli config for given profile string
        cluster_id (str): Cluster ID
        cluster_name (str): Cluster name
        local_env (str): Name of the local conda environment
        python_path (str): Remote python path to be used for kernel
    """
    from remote_ikernel.manage import show_kernel, add_kernel
    from remote_ikernel.compat import kernelspec as ks

    print("   => Creating kernel specification for profile '%s'" % profile)
    env = "DBJL_PROFILE=%s DBJL_HOST=%s DBJL_CLUSTER=%s" % (profile, host, cluster_id)
    if organisation is not None:
        env += " DBJL_ORG=%s" % organisation
    kernel_cmd = "sudo -H %s %s/python -m ipykernel -f {connection_file}" % (env, python_path)

    if cluster_name.replace(" ", "_") == local_env:
        name = "%s:%s" % (profile, cluster_name)
    else:
        name = "%s:%s (%s)" % (profile, cluster_name, local_env)

    add_kernel(
        "ssh",
        name=name,
        kernel_cmd=kernel_cmd,
        language="python",
        workdir="/home/ubuntu",
        host="%s:2200" % cluster_id,
        ssh_timeout="10",
        no_passwords=True,
        verbose=True)

    print("   => Kernel specification 'SSH %s %s' created or updated" % (cluster_id, name))

def remove_kernelspecs():
    km = kernelspec.KernelSpecManager()

    kernel_id = None
    while kernel_id != "done":

        remote_ikernels = {
            kernelspec.get_kernel_spec(k).display_name : k
            for k, v in kernelspec.find_kernel_specs().items() if k.startswith("rik")
        }
        if remote_ikernels == {}:
            print_ok("   => No databricklabs_jupyterlab kernel spec left")
            break

        choice = [
            inquirer.List("kernel_name",
                        message="Which kernel spec to delete (Ctrl-C to finish)?",
                        choices=list(remote_ikernels.keys()))
        ]
        answer = inquirer.prompt(choice, theme=Dark())

        if answer is None:
            break

        kernel_name = answer["kernel_name"]
        if kernel_id != "done":
            answer = input("Really delete kernels spec '%s' (y/n) " % kernel_name)
            if answer.lower() == "y":
                km.remove_kernel_spec(remote_ikernels[kernel_name])



import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "databrickslabs_jupyterlab",
    version = "1.0.2-rc2",
    author = "Bernhard Walter",
    author_email = "bernhard.walter@databricks.com",
    description = ("Remote Jupyter Lab kernel for Databricks"),
    long_description_content_type='text/markdown',
    license = "Databricks License",
    keywords = "databricks jupyter jupyterlab spark",
    packages=find_packages(),
    scripts=['databrickslabs-jupyterlab'],
    install_requires=[
        'notebook==5.7.8',
        'inquirer',
        'ssh_config',
        'databricks_cli'
    ],
    data_files=[
        ('etc/jupyter/jupyter_notebook_config.d', ['databrickslabs_jupyterlab/status/etc/serverextension.json']),
    ],
    include_package_data=True,
    long_description=read('Readme.md'),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: Other/Proprietary License",
    ],
    zip_safe=False
)
